import json
from datetime import UTC, datetime, timedelta

import jwt
import requests
from requests import Response

from ...abc.authentication import DatagardenEnvironment
from .settings import (
    BEARER_KEY,
    DEFAULT_HEADER,
    REFRESH_TOKEN_URL_EXTENSION,
    REQ_TOKEN_URL_EXTENSION,
)


class AccessToken:
    _instance = None
    _tokens: dict = {}
    TOKEN_LIFE_TIME_MARGIN: int = 20
    ACCESS_TOKEN_KEY = "access"
    REFRESH_TOKEN_KEY = "refresh"

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance

    def __init__(
        self,
        environment: type[DatagardenEnvironment],
        email: str | None = None,
        password: str | None = None,
    ) -> None:
        self._environment = environment()
        self._token_payload = self._environment.credentials(email, password)
        self._token_header = DEFAULT_HEADER.copy()

    @property
    def _token_url(self) -> str:
        return self._the_datagarden_url + REQ_TOKEN_URL_EXTENSION

    @property
    def _refresh_token_url(self) -> str:
        return self._the_datagarden_url + REFRESH_TOKEN_URL_EXTENSION

    @property
    def _the_datagarden_url(self) -> str:
        return self._environment.the_datagarden_url

    def _access_token_expired(self) -> bool:
        if self._token_expiry_time:
            return datetime.now(tz=UTC) + timedelta(seconds=5) > self._token_expiry_time
        return True

    @property
    def _access_token(self) -> str:
        if not self._tokens:
            self._request_tokens()
        elif self._access_token_expired():
            self._get_refresh_token()
        return self._tokens.get(self.ACCESS_TOKEN_KEY, "")

    @property
    def header_with_access_token(self) -> dict[str, str]:
        header = DEFAULT_HEADER.copy()
        header["Authorization"] = BEARER_KEY + self._access_token
        return header

    def _request_tokens(self):
        response = requests.request(
            method="POST",
            url=self._token_url,
            headers=self._token_header,
            data=json.dumps(self._token_payload),
        )
        if not response.status_code == 200:
            print("Token request failed and returned error(s): ")
            print("    Errorr : ", response.json().get("detail", "No error details provided"))
            quit()

        self._tokens = self._get_response_data(response)
        self._set_token_expiry_time()

    def _get_refresh_token(self):
        response = requests.request(
            method="POST",
            url=self._refresh_token_url,
            headers=self._token_header,
            data=json.dumps(self._refresh_payload()),
        )
        if not response.status_code == 200:
            raise ValueError("Token request failed and returned error: " f"{response.text}")
        self._tokens = self._get_response_data(response)
        self._set_token_expiry_time()

    def _refresh_payload(self) -> dict:
        refresh_token = self._tokens.get(self.REFRESH_TOKEN_KEY, "")
        return {"refresh": refresh_token}

    def _set_token_expiry_time(self):
        if not self._tokens:
            return
        access_token = self._tokens.get(self.ACCESS_TOKEN_KEY, "")
        if not access_token:
            raise ValueError("Access token not found in response")
        decoded_token = jwt.decode(access_token, options={"verify_signature": False})
        exp_time_stamp = decoded_token["exp"]
        self._token_expiry_time = datetime.fromtimestamp(timestamp=exp_time_stamp, tz=UTC)

    def _get_response_data(self, response: Response) -> dict[str, str]:
        return json.loads(response.text)



"""
Authentication settings for The DataGarden API.

This module contains constants and default values used for authentication
and API communication with The DataGarden service.

Constants:
    REQ_TOKEN_URL_EXTENSION (str): URL extension for requesting a token.
    REFRESH_TOKEN_URL_EXTENSION (str): URL extension for refreshing a token.
    HTTPS_PREFIX (str): Prefix for HTTPS URLs.
    BEARER_KEY (str): Prefix for Bearer token authentication.
    API_EXTENSION (str): Extension for API endpoints (currently empty).
    DEFAULT_HEADER (dict): Default headers for API requests.

"""

from decouple import config

from the_datagarden.version import __version__

INCLUDE_STATISTIC_PARAM = {"include_statistics": "y"}

STATISTICS_URL_EXTENSION = "statistics/"

SHOW_REQ_DETAIL = config("SHOW_REQ_DETAIL", default=False, cast=bool)
REQ_TOKEN_URL_EXTENSION = "user/token/"
REFRESH_TOKEN_URL_EXTENSION = "user/token/refresh/"

REGISTRATION_URL_EXTENSION = "user/enroll/"

HTTPS_PREFIX = "https://"
BEARER_KEY = "Bearer "
API_EXTENSION = ""

DEFAULT_HEADER: dict[str, str] = {
    "accept": "application/json",
    "accept-language": "en,nl-NL;q=0.9,nl;q=0.8,sv;q=0.7",
    "content-type": "application/json",
    "X-The-Datagarden-Version": "The_datagarden_pypi_package_version__" + __version__,
}


class URLExtension:
    ROOT = ""
    PULSE = "pulse/"
    WORLD = "world/"
    WORLD_DATA = "world/regional_data/"
    CONTINENTS = "continent/"
    CONTINENT = "continent/"
    COUNTRIES = "country/"
    COUNTRY = "country/"


class DynamicEndpointCategories:
    CONTINENTS = "continents"
    COUNTRIES = "countries"



import pandas as pd
import polars as pl
from datagarden_models import DataGardenModel, DatagardenModels, DataGardenSubModel, RegionalDataStats
from datagarden_models.models.base.legend import Legend
from pydantic import BaseModel

from the_datagarden.api.base import BaseApi

UNIQUE_FIELDS = [
    "region_type",
    "un_region_code",
    "iso_cc_2",
    "local_region_code",
    "local_region_code_type",
    "region_level",
    "period",
    "period_type",
    "source_name",
]
DEFAULT_COLUMNS_TO_EXCLUDE = [
    "datagarden_model_version",
    "name",
    "region_type",
    "un_region_code",
    "iso_cc_2",
    "local_region_code",
    "local_region_code_type",
    "parent_region_code",
    "parent_region_code_type",
    "parent_region_type",
    "region_level",
    "source_name",
    "data_model_name",
    "period",
    "period_type",
]


class RegionalDataRecord(BaseModel):
    name: str | None = None
    region_type: str | None = None
    un_region_code: str | None = None
    iso_cc_2: str | None = None
    local_region_code: str | None = None
    local_region_code_type: str | None = None
    parent_region_code: str | None = None
    parent_region_code_type: str | None = None
    parent_region_type: str | None = None
    region_level: int = 0
    source_name: str | None = None
    period: str | None = None
    period_type: str | None = None
    data_model_name: str | None = None
    model: DataGardenSubModel

    def record_hash(self) -> str:
        hash_str = ".".join([str(getattr(self, key)) for key in sorted(UNIQUE_FIELDS)])
        return str(hash(hash_str))

    def __str__(self):
        return (
            f"RegionalDataRecord: {self.name} ({self.data_model_name} for {self.period}, {self.period_type})"
        )

    @property
    def datgarden_model_class(self) -> type[DataGardenModel]:
        return self.model.__class__

    def record_for_sub_model(self, sub_model_name: str) -> "RegionalDataRecord":
        if sub_model_name not in self.datgarden_model_class.legends().sub_model_names:
            raise ValueError(f"Sub model `{sub_model_name}` not found in {self.datgarden_model_class}")
        child_record = self.model_dump()
        child_record["data_model_name"] = sub_model_name
        child_record["model"] = getattr(self.model, sub_model_name)
        return RegionalDataRecord(**child_record)


class TheDataGardenRegionalDataModel:
    """
    Model to hold response data from the The Data Garden API Regional Data endpoint.

    The model hold a list of regional_data records containg a regional data model
    for the region for a specific set op sources, periods and period types.

    The data can be converted to Polars and Pandas dataframes by the following
    methods:
    - to_polars(model_convertors: dict | None = None) -> pl.DataFrame
        model_convertors dict will be used to covert specifc model fields to dataframe
        columns.
    - full_model_to_polars() -> pl.DataFrame

    For pandas dataframes you can use the same methods:
    - to_pandas(model_convertors: dict | None = None) -> pd.DataFrame
    - full_model_to_pandas() -> pd.DataFrame
    """

    def __init__(
        self,
        api: "BaseApi",
        model_name: str,
        region_url: str,
        meta_data: BaseModel,
        is_sub_model: bool = False,
        model: type[DataGardenSubModel] | None = None,
    ):
        self._api: BaseApi = api
        self._model_name: str = model_name
        self._region_url: str = region_url
        self._request_params_hashes: list[str] = []
        self._data_records: dict[str, RegionalDataRecord] = {}
        self.meta_data: BaseModel = meta_data
        self._model: DataGardenModel = model or getattr(DatagardenModels, model_name.upper())
        self._is_sub_model: bool = is_sub_model

    def __str__(self):
        return f"TheDataGardenRegionalDataModel : {self._model_name} : (count={len(self._data_records)})"

    def __repr__(self):
        return self.__str__()

    def __call__(self, **kwargs) -> "TheDataGardenRegionalDataModel":
        if self._is_sub_model:
            raise TypeError(
                "Sub model data cannot be used to retrieve data. "
                "Use the main model data object to make calls to The-Datagarden API"
            )
        request_hash = self.request_hash(**kwargs)
        if request_hash not in self._request_params_hashes:
            regional_data = self.regional_paginated_data_from_api(**kwargs)
            if regional_data:
                self.set_items(regional_data)
                self._request_params_hashes.append(request_hash)
        return self

    def __getattr__(self, attribute: str) -> "TheDataGardenRegionalDataModel":
        if attribute not in self._model.legends().sub_model_names:
            raise ValueError(f"Attribute {attribute} is not a sub-model of {self._model_name}")
        sub_model = getattr(self._model.legends(), attribute).model
        regional_data_for_attribute = TheDataGardenRegionalDataModel(
            api=self._api,
            model_name=attribute,
            region_url=self._region_url,
            meta_data=self.meta_data,
            is_sub_model=True,
            model=sub_model,
        )
        regional_data_for_attribute._data_records = {
            key: value.record_for_sub_model(attribute) for key, value in self._data_records.items()
        }
        return regional_data_for_attribute

    @property
    def model_attributes(self) -> list[str]:
        return self._model.legends().attributes

    def model_attribute_legend(self, attribute: str) -> Legend:
        return getattr(self._model.legends(), attribute)

    def request_hash(self, **kwargs) -> str:
        sorted_items = sorted(kwargs.items())
        hash_str = ",".join(f"{k}:{v}" for k, v in sorted_items)
        return str(hash(hash_str))

    def _response_has_next_page(self, model_data_resp: dict) -> bool:
        pagination = model_data_resp.get("pagination", None)
        if not pagination:
            return False
        return pagination.get("next_page", None) is not None

    def _next_page_pagination(self, model_data_resp: dict) -> dict | None:
        pagination = model_data_resp.pop("pagination", None)
        if not pagination:
            return None
        next_page = pagination.get("next_page", None)
        if not next_page:
            return None
        return {"page": next_page}

    def regional_paginated_data_from_api(self, **kwargs) -> dict:
        model_data_resp = self.regional_data_from_api(**kwargs)
        if not model_data_resp:
            return {}
        while self._response_has_next_page(model_data_resp):
            next_page_pagination = self._next_page_pagination(model_data_resp)
            if next_page_pagination:
                next_page_resp = self.regional_data_from_api(pagination=next_page_pagination, **kwargs)
                if next_page_resp:
                    model_data_resp["data_by_region"].extend(next_page_resp["data_by_region"])
                    model_data_resp["pagination"] = next_page_resp["pagination"]

        return model_data_resp

    def regional_data_from_api(self, **kwargs) -> dict:
        model_data_resp = self._api.retrieve_from_api(
            url_extension=self._region_url + "regional_data/",
            method="POST",
            payload={"model": self._model_name, **kwargs},
        )
        if model_data_resp:
            return model_data_resp.json()
        return {}

    def set_items(self, data: dict):
        for regional_data in data["data_by_region"]:
            base_items = {
                "name": regional_data.get("region_name", None),
                "region_type": regional_data.get("region_type", None),
                "un_region_code": regional_data.get("un_region_code", None),
                "iso_cc_2": regional_data.get("iso_cc_2", None),
                "local_region_code": regional_data.get("local_region_code", None),
                "local_region_code_type": regional_data.get("local_region_code_type", None),
                "parent_region_code": regional_data.get("parent_region_code", None),
                "parent_region_code_type": regional_data.get("parent_region_code_type", None),
                "parent_region_type": regional_data.get("parent_region_type", None),
                "region_level": regional_data.get("region_level", 0),
            }
            data_for_region = regional_data["data_objects_for_region"]
            data_records = [
                RegionalDataRecord(**base_items, **self._record_items(data_obj))
                for data_obj in data_for_region
            ]
            for data_record in data_records:
                self._data_records.update({data_record.record_hash(): data_record})

        if self._data_records:
            _, first_record = list(self._data_records.items())[0]
            model_name = first_record.data_model_name
            if not model_name:
                raise ValueError("data_model_name is required")
            self._model_name = model_name

    def _record_items(self, data: dict):
        model_name = data.get("data_type", None)
        if not model_name:
            raise ValueError("data_model_name is required")

        model = getattr(DatagardenModels, model_name.upper())
        if not model:
            raise ValueError(f"model {model_name} not found in DatagardenModels")
        return {
            "source_name": data.get("source_name", None),
            "period": data.get("period", None),
            "period_type": data.get("period_type", None),
            "data_model_name": data.get("data_type", None),
            "model": model(**data.get("data", {})),
        }

    def to_polars(self, model_convertors: dict | None = None) -> pl.DataFrame:
        """
        Convert the data to a polars dataframe using a dictionary of model attributes to convert to columns
        """
        model_convertors = model_convertors or {}
        converted_records = []
        for record in self._data_records.values():
            model = record.model
            record_dict = record.model_dump()
            record_dict.pop("model")

            for new_col, model_attr in model_convertors.items():
                # Handle nested attributes using split by dots
                model_attr_flatten = "__flatten" in model_attr
                model_attr = model_attr.replace("__flatten", "")
                attrs = model_attr.split(".")
                value = getattr(model, attrs[0])
                for attr in attrs[1:]:
                    value = getattr(value, attr, None)
                if not value:
                    continue
                if model_attr_flatten:
                    model_data = value.model_dump() if isinstance(value, BaseModel) else value
                    record_dict.update(self.flatten_dict(model_data, {}))
                else:
                    record_dict[new_col] = value
            converted_records.append(record_dict)
        return pl.from_records(converted_records)

    def flatten_dict(self, dict_to_flatten: dict, flattened_dict: dict, prefix: str = "") -> dict:
        for key, value in dict_to_flatten.items():
            new_key = f"{prefix}.{key}" if prefix else key
            if isinstance(value, dict):
                flattened_dict.update(self.flatten_dict(value, flattened_dict, new_key))
            else:
                flattened_dict[new_key] = value

        return flattened_dict

    def full_model_to_polars(self):
        """
        Convert the data to a polars dataframe, flattening all nested dictionaries
        """
        converted_records = []
        for record in self._data_records.values():
            # Get all fields from the record excluding the modeL
            record_dict = record.model_dump(exclude={"model"})
            # Model data is added as flattened dictionary
            model_data = record.model.model_dump()
            flattened_dict = self.flatten_dict(model_data, {})
            record_dict.update(flattened_dict)
            converted_records.append(record_dict)
        return pl.from_records(converted_records)

    def to_pandas(self, model_convertors: dict | None = None) -> pd.DataFrame:
        """
        Convert the data to a pandas dataframe using a dictionary of model attributes to convert to columns
        """
        return self.to_polars(model_convertors).to_pandas()

    def full_model_to_pandas(self) -> pd.DataFrame:
        """
        Convert the data to a pandas dataframe, flattening all nested dictionaries
        """
        return self.full_model_to_polars().to_pandas()

    def __iter__(self):
        """Makes the class iterable over the values in _data_records"""
        return iter(self._data_records.values())

    def __len__(self):
        """Returns the number of records"""
        return len(self._data_records)

    @property
    def data_records(self) -> list[RegionalDataRecord]:
        return list(self._data_records.values())

    def regional_availability(self) -> dict[str, RegionalDataStats | None]:
        availability_per_region = self.meta_data.statistics_for_data_model(model_name=self._model_name)
        regional_availability = {}
        for region_type in self.meta_data.region_types:
            if region_type in availability_per_region.keys():
                regional_availability[region_type] = availability_per_region[region_type]
            else:
                regional_availability[region_type] = None
        return regional_availability

    @property
    def regions_with_model_data(self) -> list[str]:
        return [region for region in self.regional_availability() if self.regional_availability()[region]]

    def show_summary(self):
        """
        Outputs a summary of the model's structure (submodels and attributes)
        """
        self._model.legends().show_summary()

    def summary(self) -> dict:
        """
        return model's structure (submodels and attributes)
        """
        return self._model.legends().summary()

    def describe(
        self,
        include_attributes: list[str] | None = None,
        exclude_attributes: list[str] | None = None,
        filter_expr: pl.Expr | None = None,
    ) -> pl.DataFrame:
        df = self.full_model_to_polars()
        if df.is_empty():
            raise ValueError("No data loaded for this model. Data is needed to describe the model.")

        if filter_expr is not None:
            df = df.filter(filter_expr)

        if include_attributes:
            return df.select(include_attributes).describe()

        attributes_to_exclude = DEFAULT_COLUMNS_TO_EXCLUDE.copy()
        if exclude_attributes:
            attributes_to_exclude.extend(exclude_attributes)
        return df.select([col for col in df.columns if col not in attributes_to_exclude]).describe()

    def data_availability_per_attribute(
        self, include_attributes: list[str] | None = None, filter_expr: pl.Expr | None = None
    ):
        if include_attributes:
            describe_df = self.describe(include_attributes=include_attributes, filter_expr=filter_expr)
        else:
            describe_df = self.describe(
                exclude_attributes=DEFAULT_COLUMNS_TO_EXCLUDE, filter_expr=filter_expr
            )

        describe_df = describe_df.with_columns(
            pl.when(pl.col("statistic").is_in(["count", "null_count"]))
            .then(pl.all().exclude("statistic").cast(pl.Int64))
            .otherwise(pl.all().exclude("statistic"))
        )

        return describe_df

    def show_data_availability_per_attribute(
        self, include_attributes: list[str] | None = None, filter_expr: pl.Expr | None = None
    ):
        describe_df = self.data_availability_per_attribute(include_attributes, filter_expr)
        stats_by_column = {
            column: dict(
                zip(describe_df.get_column("statistic"), describe_df.get_column(column), strict=True)
            )
            for column in describe_df.columns
            if column != "statistic"
        }

        max_column_length = max(len(column) for column in stats_by_column.keys())

        for column, stats in stats_by_column.items():
            print(
                f"{column} : {" " * (max_column_length + 3 - len(column))}"
                f"{int(stats['count'] + stats['null_count'])}"
                f" of which with data: {int(stats['count'])} "
                f"({int(stats['count']) / (int(stats['count'] + stats['null_count'])) * 100:.0f}%)"
            )



import requests
from decouple import config

from the_datagarden.abc.authentication import BaseDataGardenCredentials, TheDatagardenCredentialsDict
from the_datagarden.api.authentication.settings import REGISTRATION_URL_EXTENSION


class CredentialsFromUserInput:
    def get_missing_credentials(self, the_datagarden_api_url: str) -> TheDatagardenCredentialsDict:
        print()
        print("Welcome to The Data Garden API.")
        print()
        print("  You can start using the API with an account from The-Datagarden.io.")
        print("  Please provide your credentials or create a new account.")
        print("  Check www.the-datagarden.io for more information.")
        print()
        choice = input(
            "Do you want to (1) create a new account or (2) provide existing credentials? " "Enter 1 or 2: "
        )

        if choice == "1":
            credentials = self.enroll_to_api(the_datagarden_api_url)
            if not credentials:
                quit()
            return credentials

        elif choice == "2":
            return self.provide_existing_credentials()
        else:
            print("Invalid choice. Please enter 1 or 2.")
            return self.get_missing_credentials(the_datagarden_api_url)

    def enroll_to_api(self, the_datagarden_api_url: str) -> TheDatagardenCredentialsDict | None:
        print("Enrolling in The Data Garden API...")
        print()
        email = input("  Enter your email: ")
        password = self.get_confirmed_password()
        print()

        data = {"email": email, "password": password}
        registration_url = the_datagarden_api_url + REGISTRATION_URL_EXTENSION
        response = requests.post(registration_url, data=data)
        if response.status_code == 201:
            print("Successfully enrolled in The Data Garden API.")
            return TheDatagardenCredentialsDict(
                email=email,
                password=password,
            )
        else:
            print("Enrollment failed with error(s):")
            for field, message_list in response.json().items():
                print(f"    {field}: {message_list[0]}")

        return None

    def get_confirmed_password(self) -> str:
        while True:
            password = input("  Enter your password: ")
            confirm_password = input("  Confirm your password: ")

            if password == confirm_password:
                return password
            else:
                print("Passwords do not match. Please try again.")

    def provide_existing_credentials(self) -> TheDatagardenCredentialsDict:
        print("Please provide your existing credentials...")
        print()
        email = input("  Enter your email: ")
        password = input("  Enter your password: ")
        return TheDatagardenCredentialsDict(
            email=email,
            password=password,
        )

    def credentials(self, the_datagarden_api_url: str) -> TheDatagardenCredentialsDict:
        return self.get_missing_credentials(the_datagarden_api_url)


class TheDataGardenCredentials(BaseDataGardenCredentials):
    """
    Manages credentials for The Data Garden API in a production environment.

    This class handles the retrieval of user credentials (email and password) for
    authenticating with The Data Garden API. It first attempts to fetch credentials
    from environment variables. If not found, it prompts the user for input.

    Attributes:
        ENV_EMAIL_KEY (str): Environment variable key for the user's email.
        ENV_PASSWORD_KEY (str): Environment variable key for the user's password.
        CREDENTIALS_FROM_USER_INPUT (CredentialsFromUserInput):
                Instance to handle user input for credentials.

    Methods:
        credentials(): Retrieves and returns the user's credentials.
    """

    ENV_EMAIL_KEY: str = "THE_DATAGARDEN_USER_EMAIL"
    ENV_PASSWORD_KEY: str = "THE_DATAGARDEN_USER_PASSWORD"
    CREDENTIALS_FROM_USER_INPUT = CredentialsFromUserInput()

    @classmethod
    def credentials(
        cls, the_datagarden_api_url: str, email: str | None = None, password: str | None = None
    ) -> TheDatagardenCredentialsDict:
        if email and password:
            return TheDatagardenCredentialsDict(
                email=email,
                password=password,
            )

        datagarden_user_email = str(config(cls.ENV_EMAIL_KEY, default="", cast=str))
        datagarden_user_password = str(config(cls.ENV_PASSWORD_KEY, default="", cast=str))
        if datagarden_user_email and datagarden_user_password:
            return TheDatagardenCredentialsDict(
                email=datagarden_user_email,
                password=datagarden_user_password,
            )

        return cls.CREDENTIALS_FROM_USER_INPUT.credentials(the_datagarden_api_url)



from enum import StrEnum


class ResponseKeys(StrEnum):
    AVAILABLE_MODELS = "AVAILABLE_MODELS"
    STATISTICS = "STATISTICS"


class RegionKeys(StrEnum): ...


class ContinentKeys(RegionKeys):
    AVAILABLE_MODELS = "available_data_on_continent_level"
    STATISTICS = "statistics"


class CountryKeys(RegionKeys):
    AVAILABLE_MODELS = "models_per_region_level"
    STATISTICS = "statistics"



"""
Base module for The Data Garden API.

This module provides the foundation for interacting with The Data Garden API.
It includes the base class for API interactions and imports necessary
authentication components.

Classes:
    TheDataGardenAPI: Base class for interacting with The Data Garden API.

Imports:
    DatagardenEnvironment: Abstract base class for environment authentication.
    TheDatagardenProductionEnvironment: Concrete implementation of the production
                                                                        environment.
    URLExtension: Class for handling URL extensions.
"""

from collections import defaultdict
from typing import Iterator

import requests
from requests import Response

from the_datagarden.abc.api import BaseApi
from the_datagarden.abc.authentication import DatagardenEnvironment
from the_datagarden.api.authentication import AccessToken
from the_datagarden.api.authentication.environment import TheDatagardenProductionEnvironment
from the_datagarden.api.authentication.settings import (
    DEFAULT_HEADER,
    SHOW_REQ_DETAIL,
    DynamicEndpointCategories,
    URLExtension,
)
from the_datagarden.api.regions import Continent
from the_datagarden.api.regions.country import Country


class BaseDataGardenAPI(BaseApi):
    """
    Base class for interacting with The Data Garden API.
    """

    ACCESS_TOKEN: type[AccessToken] = AccessToken
    DYNAMIC_ENDPOINTS: dict = defaultdict(dict)

    def __init__(
        self,
        environment: type[DatagardenEnvironment] | None = None,
        email: str | None = None,
        password: str | None = None,
    ):
        self._environment = environment or TheDatagardenProductionEnvironment
        self._base_url = self._environment().the_datagarden_url
        self._api_status = self._check_pulse()
        self._tokens = self.ACCESS_TOKEN(self._environment, email, password)

    def _check_pulse(self) -> bool:
        url = self._generate_url(URLExtension.PULSE)
        response = requests.request(method="GET", url=url, headers=DEFAULT_HEADER.copy())

        if response.status_code == 200:
            return True
        return False

    def _generate_url(self, url_extension: str) -> str:
        url = self._base_url + url_extension
        if url[-1] != "/":
            url += "/"
        return url

    def retrieve_from_api(
        self,
        url_extension: str,
        method: str = "GET",
        payload: dict | None = None,
        params: dict | None = None,
    ) -> Response | None:
        url = self._generate_url(url_extension)
        headers = self._tokens.header_with_access_token
        if SHOW_REQ_DETAIL:
            print(f"Request URL: {url}")
            print(f"Request headers: {headers}")
            print(f"Request method: {method}")
            print(f"Request payload: {payload}")
            print(f"Request params: {params}")
        match method:
            case "GET":
                response = requests.get(url, params=params, headers=headers)
            case "POST":
                response = requests.post(url, json=payload, headers=headers)
            case _:
                raise ValueError(f"Invalid method: {method}")

        return self._response_handler(response)

    def _response_handler(self, response: requests.Response) -> Response | None:
        if response.status_code == 200:
            return response
        else:
            response_dict = response.json()
            for k, v in response_dict.items():
                print(k)
                print(v)
            return None

    def _get_next_page(self, response: requests.Response) -> requests.Response | None:
        next_url = response.json().get("next")
        if not next_url:
            return None

        # Determine the original request method
        original_method = response.request.method

        headers = self._tokens.header_with_access_token

        if original_method == "GET":
            return requests.get(next_url, headers=headers)
        elif original_method == "POST":
            # For POST requests, we need to preserve the original payload
            original_payload = response.request.body
            return requests.post(next_url, data=original_payload, headers=headers)
        else:
            raise ValueError(f"Unsupported method for pagination: {original_method}")

    def _records_from_paginated_api_response(self, response: requests.Response | None) -> Iterator[dict]:
        while response:
            for record in response.json()["results"]:
                yield record
            response = self._get_next_page(response)

    def _create_url_extension(self, url_extensions: list[str]) -> str:
        url = "/".join(url_extensions).lower().replace(" ", "-")
        if url_extensions[-1] == "/":
            return url
        return url + "/"


class TheDataGardenAPI(BaseDataGardenAPI):
    _instance = None
    _initialized = False

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance

    def __init__(
        self,
        environment: type[DatagardenEnvironment] | None = None,
        email: str | None = None,
        password: str | None = None,
    ):
        if not self._initialized:
            super().__init__(environment, email, password)
            self._setup_continents()
            self._setup_countries()
            self.__class__._initialized = True

    def __getattr__(self, attr: str):
        for _, endpoints in self.DYNAMIC_ENDPOINTS.items():
            if attr.lower() in endpoints:
                return endpoints[attr.lower()]

        raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{attr}'")

    def world(self):
        response = self.retrieve_from_api(URLExtension.WORLD)
        return response.json()

    def continents(self, include_details: bool = False) -> list[str] | dict:
        continents = (
            self.DYNAMIC_ENDPOINTS.get(DynamicEndpointCategories.CONTINENTS, None) or self._setup_continents()
        )
        if not include_details:
            return continents.keys()
        return continents

    def countries(self, include_details: bool = False) -> list[str] | dict:
        countries = (
            self.DYNAMIC_ENDPOINTS.get(DynamicEndpointCategories.COUNTRIES, None) or self._setup_countries()
        )
        if not include_details:
            return countries.keys()
        return countries

    def _setup_continents(self):
        if not self.DYNAMIC_ENDPOINTS.get(DynamicEndpointCategories.CONTINENTS, None):
            continents = self.retrieve_from_api(URLExtension.CONTINENTS)
            for continent in self._records_from_paginated_api_response(continents):
                continent_method_name = continent["name"].lower().replace(" ", "_")
                self.DYNAMIC_ENDPOINTS[DynamicEndpointCategories.CONTINENTS].update(
                    {
                        continent_method_name: Continent(
                            url=self._create_url_extension([URLExtension.CONTINENT + continent["name"]]),
                            api=self,
                            name=continent["name"].lower(),
                        ),
                    }
                )

        return self.DYNAMIC_ENDPOINTS[DynamicEndpointCategories.CONTINENTS]

    def _setup_countries(self):
        if not self.DYNAMIC_ENDPOINTS.get(DynamicEndpointCategories.COUNTRIES, None):
            countries = self.retrieve_from_api(URLExtension.COUNTRIES)
            if not countries:
                return None
            for country in self._records_from_paginated_api_response(countries):
                country_method_name = country["name"].lower().replace(" ", "_")
                country_code = country["iso_cc_2"].lower()
                continent = country["parent_region"].lower()
                country = Country(
                    url=self._create_url_extension([URLExtension.COUNTRY + country["name"]]),
                    api=self,
                    name=country["name"],
                    continent=continent,
                )
                self.DYNAMIC_ENDPOINTS[DynamicEndpointCategories.COUNTRIES].update(
                    {
                        country_method_name: country,
                        country_code: country,
                    }
                )

        return self.DYNAMIC_ENDPOINTS[DynamicEndpointCategories.COUNTRIES]



import click


@click.command()
@click.argument("arg1", required=False)
def main(arg1):
    """Command-line tool for my_package_name."""
    print(f"Received arguments: arg1={arg1}")
    print("Running your command...")



from the_datagarden.abc import DatagardenEnvironment

from ..credentials import TheDataGardenCredentials


class TheDatagardenProductionEnvironment(DatagardenEnvironment):
    CREDENTIALS = TheDataGardenCredentials
    THE_DATAGARDEN_URL = "https://api.the-datagarden.io"


class TheDatagardenLocalEnvironment(DatagardenEnvironment):
    CREDENTIALS = TheDataGardenCredentials
    THE_DATAGARDEN_URL = "http://127.0.0.1:8000"



from .api.authentication.environment import TheDatagardenLocalEnvironment
from .api.base import TheDataGardenAPI, TheDatagardenProductionEnvironment

__all__ = [
    "TheDataGardenAPI",
    "TheDatagardenProductionEnvironment",
    "TheDatagardenLocalEnvironment",
]



from typing import Any

import pandas as pd
import polars as pl
from pydantic import BaseModel

from the_datagarden.api.base import BaseApi

GEJSON_UNIQUE_FIELDS = [
    "region_type",
    "iso_cc_2",
    "local_region_code",
    "local_region_code_type",
    "region_level",
]


class Properties(BaseModel):
    name: str
    region_level: int
    region_type: str
    iso_cc_2: str
    local_region_code: str
    local_region_code_type: str


class Geometry(BaseModel):
    type: str
    coordinates: list


class Feature(BaseModel):
    type: str = "Feature"
    properties: Properties
    geometry: Geometry


class RegionGeoJSONDataRecord(BaseModel):
    name: str | None = None
    region_type: str | None = None
    iso_cc_2: str | None = None
    local_region_code: str | None = None
    local_region_code_type: str | None = None
    region_level: int = 0
    feature: Feature

    def record_hash(self) -> str:
        hash_str = ".".join([str(getattr(self, key)) for key in sorted(GEJSON_UNIQUE_FIELDS)])
        return str(hash(hash_str))

    def __str__(self):
        return f"RegionGeoJSONDataRecord: {self.name} ({self.region_type} for {self.local_region_code})"


class TheDataGardenRegionGeoJSONModel:
    """
    Model to hold response data from the The Data Garden API Region GeoJSON endpoint.

    The model hold a list of regional_data records containg a regional data model
    for the region for a specific set op sources, periods and period types.

    The data can be converted to Polars and Pandas dataframes by the following
    methods:
    - to_polars(model_convertors: dict | None = None) -> pl.DataFrame
        model_convertors dict will be used to covert specifc model fields to dataframe
        columns.
    - full_model_to_polars() -> pl.DataFrame

    For pandas dataframes you can use the same methods:
    - to_pandas(model_convertors: dict | None = None) -> pd.DataFrame
    - full_model_to_pandas() -> pd.DataFrame
    """

    def __init__(self, api: "BaseApi", region_url: str):
        self._api: BaseApi = api
        self._region_url: str = region_url
        self._levels_requested: list[int] = []
        self._geojson_records: dict[str, RegionGeoJSONDataRecord] = {}

    def __str__(self):
        return f"TheDataGardenRegionGeoJSONModel : GeoJSON : (count={len(self._geojson_records)})"

    def __repr__(self):
        return self.__str__()

    def __call__(self, region_level: int = 0) -> "TheDataGardenRegionGeoJSONModel":
        if region_level not in self._levels_requested:
            features = self.geojson_paginated_data_from_api(region_level=region_level)
            if features:
                self.set_items(features)
                self._levels_requested.append(region_level)
        return self

    def _response_has_next_page(self, model_data_resp: dict) -> bool:
        pagination = model_data_resp.get("pagination", None)
        if not pagination:
            return False
        return pagination.get("next_page", None) is not None

    def _next_page_pagination(self, model_data_resp: dict) -> dict | None:
        pagination = model_data_resp.pop("pagination", None)
        if not pagination:
            return None
        next_page = pagination.get("next_page", None)
        if not next_page:
            return None
        return {"page": next_page}

    def geojson_paginated_data_from_api(self, region_level: int):
        geojson_data_resp = self.geojson_data_from_api(region_level=region_level)
        while geojson_data_resp and self._response_has_next_page(geojson_data_resp):
            next_page_pagination = self._next_page_pagination(geojson_data_resp)
            if next_page_pagination:
                next_page_resp = self.geojson_data_from_api(
                    pagination=next_page_pagination, region_level=region_level
                )
                if next_page_resp:
                    geojson_data_resp["features"].extend(next_page_resp["features"])
                    geojson_data_resp["pagination"] = next_page_resp["pagination"]

        return geojson_data_resp

    def geojson_data_from_api(
        self, region_level: int, pagination: dict[str, str] | None = None
    ) -> dict | None:
        payload: dict[str, Any] = {"region_level": region_level}
        if pagination:
            payload = payload | {"pagination": pagination}
        geojson_data_resp = self._api.retrieve_from_api(
            url_extension=self._region_url + "geojson/",
            method="POST",
            payload=payload,
        )
        if geojson_data_resp:
            return geojson_data_resp.json()
        return None

    def set_items(self, data: dict):
        for feature in data["features"]:
            feature = Feature(**feature)
            data_record_items = {
                "name": feature.properties.name,
                "region_type": feature.properties.region_type,
                "iso_cc_2": feature.properties.iso_cc_2,
                "local_region_code": feature.properties.local_region_code,
                "local_region_code_type": feature.properties.local_region_code_type,
                "region_level": feature.properties.region_level,
                "feature": feature,
            }
            data_record = RegionGeoJSONDataRecord(**data_record_items)
            self._geojson_records.update({data_record.record_hash(): data_record})

    def to_polars(self) -> pl.DataFrame:
        """
        Convert the data to a polars dataframe using a dictionary of model attributes to convert to columns
        """
        converted_records = []
        for record in self._geojson_records.values():
            record_dict = record.model_dump()
            converted_records.append(record_dict)
        return pl.from_records(converted_records)

    def to_pandas(self) -> pd.DataFrame:
        """
        Convert the data to a pandas dataframe
        """
        return self.to_polars().to_pandas()

    def __iter__(self):
        """Makes the class iterable over the values in _data_records"""
        return iter(self._geojson_records.values())

    def __len__(self):
        """Returns the number of records"""
        return len(self._geojson_records)

    @property
    def geojson_records(self) -> list[RegionGeoJSONDataRecord]:
        return list(self._geojson_records.values())



from abc import ABC, abstractmethod
from typing import TypedDict


class TheDatagardenCredentialsDict(TypedDict):
    email: str
    password: str


class BaseDataGardenCredentials(ABC):
    """Protocol for the datagarden credentials"""

    @classmethod
    @abstractmethod
    def credentials(
        cls, the_datagarden_api_url: str, email: str | None = None, password: str | None = None
    ) -> TheDatagardenCredentialsDict: ...


class DatagardenEnvironment(ABC):
    """Protocol for the datagarden environment"""

    CREDENTIALS: type[BaseDataGardenCredentials]
    THE_DATAGARDEN_URL: str
    ECHO_INIT: bool = True

    def __init__(self):
        if self.ECHO_INIT:
            print("Initializing :", self.__class__.__name__)
            print("At :", self.the_datagarden_url)

    def credentials(
        self, email: str | None = None, password: str | None = None
    ) -> TheDatagardenCredentialsDict:
        return self.CREDENTIALS.credentials(self.the_datagarden_url, email, password)

    @property
    def the_datagarden_url(self) -> str:
        url = self.THE_DATAGARDEN_URL
        if url[-1] != "/":
            return url + "/"
        return url



from .geojson import RegionGeoJSONDataRecord, TheDataGardenRegionGeoJSONModel
from .regional_data import RegionalDataRecord, TheDataGardenRegionalDataModel

__all__ = [
    "RegionalDataRecord",
    "TheDataGardenRegionalDataModel",
    "RegionGeoJSONDataRecord",
    "TheDataGardenRegionGeoJSONModel",
]



from datagarden_models import ContinentStats

from .base import Region
from .base.settings import ContinentKeys


class Continent(Region):
    KEYS = ContinentKeys
    REGION_STATS_MODEL = ContinentStats



from abc import ABC, abstractmethod

from requests import Response

from the_datagarden.abc.authentication import DatagardenEnvironment


class BaseApi(ABC):
    @abstractmethod
    def __init__(self, environment: type[DatagardenEnvironment] | None = None): ...

    @abstractmethod
    def retrieve_from_api(
        self,
        url_extension: str,
        method: str = "GET",
        payload: dict | None = None,
        params: dict | None = None,
    ) -> Response | None: ...



from the_datagarden.cli import main


def test_main():
    # Call main with standalone_mode=False to prevent sys.exit()
    result = main.main([], standalone_mode=False)
    assert result is None  # Or whatever return value you expect


def test_main_with_arg():
    result = main.main(["test"], standalone_mode=False)
    assert result is None



from .authentication import AccessToken
from .authentication.environment import TheDatagardenProductionEnvironment
from .base import TheDataGardenAPI

__all__ = ["AccessToken", "TheDatagardenProductionEnvironment", "TheDataGardenAPI"]



from datetime import datetime
from enum import StrEnum
from typing import Literal

from pydantic import BaseModel

from the_datagarden.api.authentication.settings import STATISTICS_URL_EXTENSION
from the_datagarden.api.base import BaseApi
from the_datagarden.models import TheDataGardenRegionalDataModel, TheDataGardenRegionGeoJSONModel

from .settings import ResponseKeys


class PeriodTypes:
    """Choice class for periodtype used in most data classes"""

    YEAR = "Y"
    QUARTER = "Q"
    MONTH = "M"
    WEEK = "W"
    DAY = "D"
    HOUR = "H"


PeriodType = Literal["Y", "Q", "M", "W", "D", "H"]


class RegionParams(BaseModel):
    models: list[str] | None = None
    source: list[str] | None = None
    period_type: PeriodType = "Y"
    period_from: datetime | None = None
    period_to: datetime | None = None
    region_type: str | None = None
    descendant_level: int = 0


class Region:
    """
    A region in The Data Garden.
    """

    REGION_STATS_MODEL: type[BaseModel]
    _region_stats: BaseModel | None = None

    KEYS: type[StrEnum]

    def __repr__(self):
        return f"{self.__class__.__name__} : {self._name}"

    def __init__(self, url: str, api: BaseApi, name: str, continent: str | None = None):
        self._region_url = url
        self._api = api
        self._available_models: dict = {}
        self._model_data_storage: dict[str, TheDataGardenRegionalDataModel] = {}
        self._geojsons = TheDataGardenRegionGeoJSONModel(api=api, region_url=url)
        self._name = name
        self._continent = continent

    def __getattr__(self, attr: str):
        if attr in self.available_model_names:
            return self._model_data_from_storage(model_name=attr)
        if attr == "geojsons":
            return self._geojsons

        raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{attr}'")

    def _model_data_from_storage(self, model_name: str) -> TheDataGardenRegionalDataModel | None:
        stored_model_data = self._model_data_storage.get(model_name, None)
        if not stored_model_data:
            self._model_data_storage[model_name] = TheDataGardenRegionalDataModel(
                model_name=model_name, api=self._api, region_url=self._region_url, meta_data=self.meta_data
            )
            return self._model_data_storage[model_name]

        return stored_model_data

    @property
    def meta_data(self) -> BaseModel | None:
        """
        Get the region statistics info from the API.
        """
        if not self._region_stats:
            region_stats_resp = self._api.retrieve_from_api(
                url_extension=self._region_url + STATISTICS_URL_EXTENSION,
            )
            if region_stats_resp and region_stats_resp.status_code == 200:
                region_stats_resp_json = region_stats_resp.json().get(self._key(ResponseKeys.STATISTICS), {})
                self._region_stats = self.REGION_STATS_MODEL(
                    region_stats_resp_json if isinstance(region_stats_resp_json, dict) else {}
                )

        return self._region_stats

    @property
    def region_types(self) -> list[str]:
        if not self.meta_data:
            return []
        return self.meta_data.region_types

    @property
    def available_model_names(self) -> list[str]:
        if not self.meta_data:
            return []
        return self.meta_data.regional_data_models

    def _key(self, key: str) -> str:
        return getattr(self.KEYS, key)
